from typing import Dict, List, Optional

from hera.shared._base_model import BaseModel as BaseModel

from ...k8s.api.core import v1 as v1
from ...k8s.apimachinery.pkg.apis.meta import v1 as v1_1

class AMQPConsumeConfig(BaseModel):
    auto_ack: Optional[bool]
    consumer_tag: Optional[str]
    exclusive: Optional[bool]
    no_local: Optional[bool]
    no_wait: Optional[bool]

class AMQPExchangeDeclareConfig(BaseModel):
    auto_delete: Optional[bool]
    durable: Optional[bool]
    internal: Optional[bool]
    no_wait: Optional[bool]

class AMQPQueueBindConfig(BaseModel):
    no_wait: Optional[bool]

class AMQPQueueDeclareConfig(BaseModel):
    arguments: Optional[str]
    auto_delete: Optional[bool]
    durable: Optional[bool]
    exclusive: Optional[bool]
    name: Optional[str]
    no_wait: Optional[bool]

class Amount(BaseModel):
    value: Optional[str]

class BitbucketRepository(BaseModel):
    owner: Optional[str]
    repository_slug: Optional[str]

class BitbucketServerRepository(BaseModel):
    project_key: Optional[str]
    repository_slug: Optional[str]

class CatchupConfiguration(BaseModel):
    enabled: Optional[bool]
    max_duration: Optional[str]

class ConditionsResetByTime(BaseModel):
    cron: Optional[str]
    timezone: Optional[str]

class ConditionsResetCriteria(BaseModel):
    by_time: Optional[ConditionsResetByTime]

class ConfigMapPersistence(BaseModel):
    create_if_not_exist: Optional[bool]
    name: Optional[str]

class DataFilter(BaseModel):
    comparator: Optional[str]
    path: Optional[str]
    template: Optional[str]
    type: Optional[str]
    value: Optional[List[str]]

class EventDependencyTransformer(BaseModel):
    jq: Optional[str]
    script: Optional[str]

class EventPersistence(BaseModel):
    catchup: Optional[CatchupConfiguration]
    config_map: Optional[ConfigMapPersistence]

class EventSourceFilter(BaseModel):
    expression: Optional[str]

class FileArtifact(BaseModel):
    path: Optional[str]

class GitRemoteConfig(BaseModel):
    name: Optional[str]
    urls: Optional[List[str]]

class Int64OrString(BaseModel):
    int64_val: Optional[str]
    str_val: Optional[str]
    type: Optional[str]

class KafkaConsumerGroup(BaseModel):
    group_name: Optional[str]
    oldest: Optional[bool]
    rebalance_strategy: Optional[str]

class LogTrigger(BaseModel):
    interval_seconds: Optional[str]

class Metadata(BaseModel):
    annotations: Optional[Dict[str, str]]
    labels: Optional[Dict[str, str]]

class OwnedRepositories(BaseModel):
    names: Optional[List[str]]
    owner: Optional[str]

class PayloadField(BaseModel):
    name: Optional[str]
    path: Optional[str]

class RateLimit(BaseModel):
    requests_per_unit: Optional[int]
    unit: Optional[str]

class Resource(BaseModel):
    value: Optional[str]

class S3Bucket(BaseModel):
    key: Optional[str]
    name: Optional[str]

class S3Filter(BaseModel):
    prefix: Optional[str]
    suffix: Optional[str]

class Selector(BaseModel):
    key: Optional[str]
    operation: Optional[str]
    value: Optional[str]

class StatusPolicy(BaseModel):
    allow: Optional[List[int]]

class StorageGridFilter(BaseModel):
    prefix: Optional[str]
    suffix: Optional[str]

class TimeFilter(BaseModel):
    start: Optional[str]
    stop: Optional[str]

class TriggerParameterSource(BaseModel):
    context_key: Optional[str]
    context_template: Optional[str]
    data_key: Optional[str]
    data_template: Optional[str]
    dependency_name: Optional[str]
    value: Optional[str]

class URLArtifact(BaseModel):
    path: Optional[str]
    verify_cert: Optional[bool]

class WatchPathConfig(BaseModel):
    directory: Optional[str]
    path: Optional[str]
    path_regexp: Optional[str]

class AzureEventsHubEventSource(BaseModel):
    filter: Optional[EventSourceFilter]
    fqdn: Optional[str]
    hub_name: Optional[str]
    metadata: Optional[Dict[str, str]]
    shared_access_key: Optional[v1.SecretKeySelector]
    shared_access_key_name: Optional[v1.SecretKeySelector]

class Backoff(BaseModel):
    duration: Optional[Int64OrString]
    factor: Optional[Amount]
    jitter: Optional[Amount]
    steps: Optional[int]

class BasicAuth(BaseModel):
    password: Optional[v1.SecretKeySelector]
    username: Optional[v1.SecretKeySelector]

class BitbucketBasicAuth(BaseModel):
    password: Optional[v1.SecretKeySelector]
    username: Optional[v1.SecretKeySelector]

class CalendarEventSource(BaseModel):
    exclusion_dates: Optional[List[str]]
    filter: Optional[EventSourceFilter]
    interval: Optional[str]
    metadata: Optional[Dict[str, str]]
    persistence: Optional[EventPersistence]
    schedule: Optional[str]
    timezone: Optional[str]

class Condition(BaseModel):
    last_transition_time: Optional[v1_1.Time]
    message: Optional[str]
    reason: Optional[str]
    status: Optional[str]
    type: Optional[str]

class EventContext(BaseModel):
    datacontenttype: Optional[str]
    id: Optional[str]
    source: Optional[str]
    specversion: Optional[str]
    subject: Optional[str]
    time: Optional[v1_1.Time]
    type: Optional[str]

class ExprFilter(BaseModel):
    expr: Optional[str]
    fields: Optional[List[PayloadField]]

class FileEventSource(BaseModel):
    event_type: Optional[str]
    filter: Optional[EventSourceFilter]
    metadata: Optional[Dict[str, str]]
    polling: Optional[bool]
    watch_path_config: Optional[WatchPathConfig]

class GenericEventSource(BaseModel):
    auth_secret: Optional[v1.SecretKeySelector]
    config: Optional[str]
    filter: Optional[EventSourceFilter]
    insecure: Optional[bool]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    url: Optional[str]

class GitCreds(BaseModel):
    password: Optional[v1.SecretKeySelector]
    username: Optional[v1.SecretKeySelector]

class GithubAppCreds(BaseModel):
    app_id: Optional[str]
    installation_id: Optional[str]
    private_key: Optional[v1.SecretKeySelector]

class HDFSEventSource(BaseModel):
    addresses: Optional[List[str]]
    check_interval: Optional[str]
    filter: Optional[EventSourceFilter]
    hdfs_user: Optional[str]
    krb_c_cache_secret: Optional[v1.SecretKeySelector]
    krb_config_config_map: Optional[v1.ConfigMapKeySelector]
    krb_keytab_secret: Optional[v1.SecretKeySelector]
    krb_realm: Optional[str]
    krb_service_principal_name: Optional[str]
    krb_username: Optional[str]
    metadata: Optional[Dict[str, str]]
    type: Optional[str]
    watch_path_config: Optional[WatchPathConfig]

class K8SResourcePolicy(BaseModel):
    backoff: Optional[Backoff]
    error_on_backoff_timeout: Optional[bool]
    labels: Optional[Dict[str, str]]

class NATSAuth(BaseModel):
    basic: Optional[BasicAuth]
    credential: Optional[v1.SecretKeySelector]
    nkey: Optional[v1.SecretKeySelector]
    token: Optional[v1.SecretKeySelector]

class PubSubEventSource(BaseModel):
    credential_secret: Optional[v1.SecretKeySelector]
    delete_subscription_on_finish: Optional[bool]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    project_id: Optional[str]
    subscription_id: Optional[str]
    topic: Optional[str]
    topic_project_id: Optional[str]

class ResourceFilter(BaseModel):
    after_start: Optional[bool]
    created_by: Optional[v1_1.Time]
    fields: Optional[List[Selector]]
    labels: Optional[List[Selector]]
    prefix: Optional[str]

class S3Artifact(BaseModel):
    access_key: Optional[v1.SecretKeySelector]
    bucket: Optional[S3Bucket]
    endpoint: Optional[str]
    events: Optional[List[str]]
    filter: Optional[S3Filter]
    insecure: Optional[bool]
    metadata: Optional[Dict[str, str]]
    region: Optional[str]
    secret_key: Optional[v1.SecretKeySelector]

class SASLConfig(BaseModel):
    mechanism: Optional[str]
    password: Optional[v1.SecretKeySelector]
    user: Optional[v1.SecretKeySelector]

class SQSEventSource(BaseModel):
    access_key: Optional[v1.SecretKeySelector]
    dlq: Optional[bool]
    endpoint: Optional[str]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    queue: Optional[str]
    queue_account_id: Optional[str]
    region: Optional[str]
    role_arn: Optional[str]
    secret_key: Optional[v1.SecretKeySelector]
    session_token: Optional[v1.SecretKeySelector]
    wait_time_seconds: Optional[str]

class Status(BaseModel):
    conditions: Optional[List[Condition]]

class TLSConfig(BaseModel):
    ca_cert_secret: Optional[v1.SecretKeySelector]
    client_cert_secret: Optional[v1.SecretKeySelector]
    client_key_secret: Optional[v1.SecretKeySelector]
    insecure_skip_verify: Optional[bool]

class TriggerParameter(BaseModel):
    dest: Optional[str]
    operation: Optional[str]
    src: Optional[TriggerParameterSource]

class TriggerPolicy(BaseModel):
    k8s: Optional[K8SResourcePolicy]
    status: Optional[StatusPolicy]

class ValueFromSource(BaseModel):
    config_map_key_ref: Optional[v1.ConfigMapKeySelector]
    secret_key_ref: Optional[v1.SecretKeySelector]

class WebhookContext(BaseModel):
    auth_secret: Optional[v1.SecretKeySelector]
    endpoint: Optional[str]
    max_payload_size: Optional[str]
    metadata: Optional[Dict[str, str]]
    method: Optional[str]
    port: Optional[str]
    server_cert_secret: Optional[v1.SecretKeySelector]
    server_key_secret: Optional[v1.SecretKeySelector]
    url: Optional[str]

class WebhookEventSource(BaseModel):
    filter: Optional[EventSourceFilter]
    webhook_context: Optional[WebhookContext]

class AMQPEventSource(BaseModel):
    auth: Optional[BasicAuth]
    connection_backoff: Optional[Backoff]
    consume: Optional[AMQPConsumeConfig]
    exchange_declare: Optional[AMQPExchangeDeclareConfig]
    exchange_name: Optional[str]
    exchange_type: Optional[str]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    queue_bind: Optional[AMQPQueueBindConfig]
    queue_declare: Optional[AMQPQueueDeclareConfig]
    routing_key: Optional[str]
    tls: Optional[TLSConfig]
    url: Optional[str]
    url_secret: Optional[v1.SecretKeySelector]

class AWSLambdaTrigger(BaseModel):
    access_key: Optional[v1.SecretKeySelector]
    function_name: Optional[str]
    invocation_type: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    region: Optional[str]
    role_arn: Optional[str]
    secret_key: Optional[v1.SecretKeySelector]

class AzureEventHubsTrigger(BaseModel):
    fqdn: Optional[str]
    hub_name: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    shared_access_key: Optional[v1.SecretKeySelector]
    shared_access_key_name: Optional[v1.SecretKeySelector]

class BitbucketAuth(BaseModel):
    basic: Optional[BitbucketBasicAuth]
    oauth_token: Optional[v1.SecretKeySelector]

class BitbucketEventSource(BaseModel):
    auth: Optional[BitbucketAuth]
    delete_hook_on_finish: Optional[bool]
    events: Optional[List[str]]
    filter: Optional[EventSourceFilter]
    metadata: Optional[Dict[str, str]]
    owner: Optional[str]
    project_key: Optional[str]
    repositories: Optional[List[BitbucketRepository]]
    repository_slug: Optional[str]
    webhook: Optional[WebhookContext]

class BitbucketServerEventSource(BaseModel):
    access_token: Optional[v1.SecretKeySelector]
    bitbucketserver_base_url: Optional[str]
    delete_hook_on_finish: Optional[bool]
    events: Optional[List[str]]
    filter: Optional[EventSourceFilter]
    metadata: Optional[Dict[str, str]]
    project_key: Optional[str]
    repositories: Optional[List[BitbucketServerRepository]]
    repository_slug: Optional[str]
    webhook: Optional[WebhookContext]
    webhook_secret: Optional[v1.SecretKeySelector]

class CustomTrigger(BaseModel):
    cert_secret: Optional[v1.SecretKeySelector]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    secure: Optional[bool]
    server_name_override: Optional[str]
    server_url: Optional[str]
    spec: Optional[Dict[str, str]]

class EmitterEventSource(BaseModel):
    broker: Optional[str]
    channel_key: Optional[str]
    channel_name: Optional[str]
    connection_backoff: Optional[Backoff]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    password: Optional[v1.SecretKeySelector]
    tls: Optional[TLSConfig]
    username: Optional[v1.SecretKeySelector]

class EventDependencyFilter(BaseModel):
    context: Optional[EventContext]
    data: Optional[List[DataFilter]]
    data_logical_operator: Optional[str]
    expr_logical_operator: Optional[str]
    exprs: Optional[List[ExprFilter]]
    script: Optional[str]
    time: Optional[TimeFilter]

class EventSourceStatus(BaseModel):
    status: Optional[Status]

class GitArtifact(BaseModel):
    branch: Optional[str]
    clone_directory: Optional[str]
    creds: Optional[GitCreds]
    file_path: Optional[str]
    insecure_ignore_host_key: Optional[bool]
    ref: Optional[str]
    remote: Optional[GitRemoteConfig]
    ssh_key_secret: Optional[v1.SecretKeySelector]
    tag: Optional[str]
    url: Optional[str]

class GithubEventSource(BaseModel):
    active: Optional[bool]
    api_token: Optional[v1.SecretKeySelector]
    content_type: Optional[str]
    delete_hook_on_finish: Optional[bool]
    events: Optional[List[str]]
    filter: Optional[EventSourceFilter]
    github_app: Optional[GithubAppCreds]
    github_base_url: Optional[str]
    github_upload_url: Optional[str]
    id: Optional[str]
    insecure: Optional[bool]
    metadata: Optional[Dict[str, str]]
    organizations: Optional[List[str]]
    owner: Optional[str]
    repositories: Optional[List[OwnedRepositories]]
    repository: Optional[str]
    webhook: Optional[WebhookContext]
    webhook_secret: Optional[v1.SecretKeySelector]

class GitlabEventSource(BaseModel):
    access_token: Optional[v1.SecretKeySelector]
    delete_hook_on_finish: Optional[bool]
    enable_ssl_verification: Optional[bool]
    events: Optional[List[str]]
    filter: Optional[EventSourceFilter]
    gitlab_base_url: Optional[str]
    metadata: Optional[Dict[str, str]]
    project_id: Optional[str]
    projects: Optional[List[str]]
    secret_token: Optional[v1.SecretKeySelector]
    webhook: Optional[WebhookContext]

class KafkaEventSource(BaseModel):
    config: Optional[str]
    connection_backoff: Optional[Backoff]
    consumer_group: Optional[KafkaConsumerGroup]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    limit_events_per_second: Optional[str]
    metadata: Optional[Dict[str, str]]
    partition: Optional[str]
    sasl: Optional[SASLConfig]
    tls: Optional[TLSConfig]
    topic: Optional[str]
    url: Optional[str]
    version: Optional[str]

class KafkaTrigger(BaseModel):
    compress: Optional[bool]
    flush_frequency: Optional[int]
    parameters: Optional[List[TriggerParameter]]
    partition: Optional[int]
    partitioning_key: Optional[str]
    payload: Optional[List[TriggerParameter]]
    required_acks: Optional[int]
    sasl: Optional[SASLConfig]
    tls: Optional[TLSConfig]
    topic: Optional[str]
    url: Optional[str]
    version: Optional[str]

class MQTTEventSource(BaseModel):
    client_id: Optional[str]
    connection_backoff: Optional[Backoff]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    tls: Optional[TLSConfig]
    topic: Optional[str]
    url: Optional[str]

class NATSEventsSource(BaseModel):
    auth: Optional[NATSAuth]
    connection_backoff: Optional[Backoff]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    subject: Optional[str]
    tls: Optional[TLSConfig]
    url: Optional[str]

class NATSTrigger(BaseModel):
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    subject: Optional[str]
    tls: Optional[TLSConfig]
    url: Optional[str]

class NSQEventSource(BaseModel):
    channel: Optional[str]
    connection_backoff: Optional[Backoff]
    filter: Optional[EventSourceFilter]
    host_address: Optional[str]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    tls: Optional[TLSConfig]
    topic: Optional[str]

class OpenWhiskTrigger(BaseModel):
    action_name: Optional[str]
    auth_token: Optional[v1.SecretKeySelector]
    host: Optional[str]
    namespace: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    version: Optional[str]

class PulsarEventSource(BaseModel):
    auth_token_secret: Optional[v1.SecretKeySelector]
    connection_backoff: Optional[Backoff]
    filter: Optional[EventSourceFilter]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    tls: Optional[TLSConfig]
    tls_allow_insecure_connection: Optional[bool]
    tls_trust_certs_secret: Optional[v1.SecretKeySelector]
    tls_validate_hostname: Optional[bool]
    topics: Optional[List[str]]
    type: Optional[str]
    url: Optional[str]

class PulsarTrigger(BaseModel):
    auth_token_secret: Optional[v1.SecretKeySelector]
    connection_backoff: Optional[Backoff]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    tls: Optional[TLSConfig]
    tls_allow_insecure_connection: Optional[bool]
    tls_trust_certs_secret: Optional[v1.SecretKeySelector]
    tls_validate_hostname: Optional[bool]
    topic: Optional[str]
    url: Optional[str]

class RedisEventSource(BaseModel):
    channels: Optional[List[str]]
    db: Optional[int]
    filter: Optional[EventSourceFilter]
    host_address: Optional[str]
    json_body: Optional[bool]
    metadata: Optional[Dict[str, str]]
    namespace: Optional[str]
    password: Optional[v1.SecretKeySelector]
    tls: Optional[TLSConfig]
    username: Optional[str]

class RedisStreamEventSource(BaseModel):
    consumer_group: Optional[str]
    db: Optional[int]
    filter: Optional[EventSourceFilter]
    host_address: Optional[str]
    max_msg_count_per_read: Optional[int]
    metadata: Optional[Dict[str, str]]
    password: Optional[v1.SecretKeySelector]
    streams: Optional[List[str]]
    tls: Optional[TLSConfig]
    username: Optional[str]

class ResourceEventSource(BaseModel):
    event_types: Optional[List[str]]
    filter: Optional[ResourceFilter]
    group_version_resource: Optional[v1_1.GroupVersionResource]
    metadata: Optional[Dict[str, str]]
    namespace: Optional[str]

class SNSEventSource(BaseModel):
    access_key: Optional[v1.SecretKeySelector]
    endpoint: Optional[str]
    filter: Optional[EventSourceFilter]
    metadata: Optional[Dict[str, str]]
    region: Optional[str]
    role_arn: Optional[str]
    secret_key: Optional[v1.SecretKeySelector]
    topic_arn: Optional[str]
    validate_signature: Optional[bool]
    webhook: Optional[WebhookContext]

class SecureHeader(BaseModel):
    name: Optional[str]
    value_from: Optional[ValueFromSource]

class SensorStatus(BaseModel):
    status: Optional[Status]

class Service(BaseModel):
    cluster_ip: Optional[str]
    ports: Optional[List[v1.ServicePort]]

class SlackEventSource(BaseModel):
    filter: Optional[EventSourceFilter]
    metadata: Optional[Dict[str, str]]
    signing_secret: Optional[v1.SecretKeySelector]
    token: Optional[v1.SecretKeySelector]
    webhook: Optional[WebhookContext]

class SlackTrigger(BaseModel):
    channel: Optional[str]
    message: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    slack_token: Optional[v1.SecretKeySelector]

class StorageGridEventSource(BaseModel):
    api_url: Optional[str]
    auth_token: Optional[v1.SecretKeySelector]
    bucket: Optional[str]
    events: Optional[List[str]]
    filter: Optional[StorageGridFilter]
    metadata: Optional[Dict[str, str]]
    region: Optional[str]
    topic_arn: Optional[str]
    webhook: Optional[WebhookContext]

class StripeEventSource(BaseModel):
    api_key: Optional[v1.SecretKeySelector]
    create_webhook: Optional[bool]
    event_filter: Optional[List[str]]
    metadata: Optional[Dict[str, str]]
    webhook: Optional[WebhookContext]

class ArtifactLocation(BaseModel):
    configmap: Optional[v1.ConfigMapKeySelector]
    file: Optional[FileArtifact]
    git: Optional[GitArtifact]
    inline: Optional[str]
    resource: Optional[Resource]
    s3: Optional[S3Artifact]
    url: Optional[URLArtifact]

class EventDependency(BaseModel):
    event_name: Optional[str]
    event_source_name: Optional[str]
    filters: Optional[EventDependencyFilter]
    filters_logical_operator: Optional[str]
    name: Optional[str]
    transform: Optional[EventDependencyTransformer]

class HTTPTrigger(BaseModel):
    basic_auth: Optional[BasicAuth]
    headers: Optional[Dict[str, str]]
    method: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    payload: Optional[List[TriggerParameter]]
    secure_headers: Optional[List[SecureHeader]]
    timeout: Optional[str]
    tls: Optional[TLSConfig]
    url: Optional[str]

class StandardK8STrigger(BaseModel):
    live_object: Optional[bool]
    operation: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    patch_strategy: Optional[str]
    source: Optional[ArtifactLocation]

class ArgoWorkflowTrigger(BaseModel):
    args: Optional[List[str]]
    operation: Optional[str]
    parameters: Optional[List[TriggerParameter]]
    source: Optional[ArtifactLocation]

class TriggerTemplate(BaseModel):
    argo_workflow: Optional[ArgoWorkflowTrigger]
    aws_lambda: Optional[AWSLambdaTrigger]
    azure_event_hubs: Optional[AzureEventHubsTrigger]
    conditions: Optional[str]
    conditions_reset: Optional[List[ConditionsResetCriteria]]
    custom: Optional[CustomTrigger]
    http: Optional[HTTPTrigger]
    k8s: Optional[StandardK8STrigger]
    kafka: Optional[KafkaTrigger]
    log: Optional[LogTrigger]
    name: Optional[str]
    nats: Optional[NATSTrigger]
    open_whisk: Optional[OpenWhiskTrigger]
    pulsar: Optional[PulsarTrigger]
    slack: Optional[SlackTrigger]

class Template(BaseModel):
    affinity: Optional[v1.Affinity]
    container: Optional[v1.Container]
    image_pull_secrets: Optional[List[v1.LocalObjectReference]]
    metadata: Optional[Metadata]
    node_selector: Optional[Dict[str, str]]
    priority: Optional[int]
    priority_class_name: Optional[str]
    security_context: Optional[v1.PodSecurityContext]
    service_account_name: Optional[str]
    tolerations: Optional[List[v1.Toleration]]
    volumes: Optional[List[v1.Volume]]

class Trigger(BaseModel):
    parameters: Optional[List[TriggerParameter]]
    policy: Optional[TriggerPolicy]
    rate_limit: Optional[RateLimit]
    retry_strategy: Optional[Backoff]
    template: Optional[TriggerTemplate]

class EventSourceSpec(BaseModel):
    amqp: Optional[Dict[str, AMQPEventSource]]
    azure_events_hub: Optional[Dict[str, AzureEventsHubEventSource]]
    bitbucket: Optional[Dict[str, BitbucketEventSource]]
    bitbucketserver: Optional[Dict[str, BitbucketServerEventSource]]
    calendar: Optional[Dict[str, CalendarEventSource]]
    emitter: Optional[Dict[str, EmitterEventSource]]
    event_bus_name: Optional[str]
    file: Optional[Dict[str, FileEventSource]]
    generic: Optional[Dict[str, GenericEventSource]]
    github: Optional[Dict[str, GithubEventSource]]
    gitlab: Optional[Dict[str, GitlabEventSource]]
    hdfs: Optional[Dict[str, HDFSEventSource]]
    kafka: Optional[Dict[str, KafkaEventSource]]
    minio: Optional[Dict[str, S3Artifact]]
    mqtt: Optional[Dict[str, MQTTEventSource]]
    nats: Optional[Dict[str, NATSEventsSource]]
    nsq: Optional[Dict[str, NSQEventSource]]
    pub_sub: Optional[Dict[str, PubSubEventSource]]
    pulsar: Optional[Dict[str, PulsarEventSource]]
    redis: Optional[Dict[str, RedisEventSource]]
    redis_stream: Optional[Dict[str, RedisStreamEventSource]]
    replicas: Optional[int]
    resource: Optional[Dict[str, ResourceEventSource]]
    service: Optional[Service]
    slack: Optional[Dict[str, SlackEventSource]]
    sns: Optional[Dict[str, SNSEventSource]]
    sqs: Optional[Dict[str, SQSEventSource]]
    storage_grid: Optional[Dict[str, StorageGridEventSource]]
    stripe: Optional[Dict[str, StripeEventSource]]
    template: Optional[Template]
    webhook: Optional[Dict[str, WebhookEventSource]]

class SensorSpec(BaseModel):
    dependencies: Optional[List[EventDependency]]
    error_on_failed_round: Optional[bool]
    event_bus_name: Optional[str]
    replicas: Optional[int]
    template: Optional[Template]
    triggers: Optional[List[Trigger]]

class EventSource(BaseModel):
    metadata: Optional[v1_1.ObjectMeta]
    spec: Optional[EventSourceSpec]
    status: Optional[EventSourceStatus]

class EventSourceList(BaseModel):
    items: Optional[List[EventSource]]
    metadata: Optional[v1_1.ListMeta]

class Sensor(BaseModel):
    metadata: Optional[v1_1.ObjectMeta]
    spec: Optional[SensorSpec]
    status: Optional[SensorStatus]

class SensorList(BaseModel):
    items: Optional[List[Sensor]]
    metadata: Optional[v1_1.ListMeta]



"""A python to expr transpiler."""

from hera.expr._node import (
    Constant as C,
    Identifier,
    Parentheses as P,
)
from hera.expr._sprig import Sprig

it = Identifier("#")
g = Identifier("")
sprig = Sprig()

__all__ = [
    "C",
    "g",
    "it",
    "P",
    "sprig",
]



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/argoproj/argo-workflows/v3.4.4/api/openapi-spec/swagger.json

from __future__ import annotations

from typing import Optional

from pydantic import Field

from hera.shared._base_model import BaseModel

from ...apimachinery.pkg.apis.meta import v1
from ...apimachinery.pkg.util import intstr


class PodDisruptionBudgetSpec(BaseModel):
    max_unavailable: Optional[intstr.IntOrString] = Field(
        None,
        alias="maxUnavailable",
        description=(
            'An eviction is allowed if at most "maxUnavailable" pods selected by "selector" are unavailable after the'
            " eviction, i.e. even in absence of the evicted pod. For example, one can prevent all voluntary evictions"
            ' by specifying 0. This is a mutually exclusive setting with "minAvailable".'
        ),
    )
    min_available: Optional[intstr.IntOrString] = Field(
        None,
        alias="minAvailable",
        description=(
            'An eviction is allowed if at least "minAvailable" pods selected by "selector" will still be available'
            " after the eviction, i.e. even in the absence of the evicted pod.  So for example you can prevent all"
            ' voluntary evictions by specifying "100%".'
        ),
    )
    selector: Optional[v1.LabelSelector] = Field(
        None,
        description=(
            "Label query over pods whose evictions are managed by the disruption budget. A null selector selects no"
            " pods. An empty selector ({}) also selects no pods, which differs from standard behavior of selecting all"
            " pods. In policy/v1, an empty selector will select all pods in the namespace."
        ),
    )



import uuid
from enum import Enum
from typing import List, Optional, Union, cast

from pydantic import root_validator, validator

from hera.workflows.models import (
    AWSElasticBlockStoreVolumeSource as _ModelAWSElasticBlockStoreVolumeSource,
    AzureDiskVolumeSource as _ModelAzureDiskVolumeSource,
    AzureFileVolumeSource as _ModelAzureFileVolumeSource,
    CephFSVolumeSource as _ModelCephFSVolumeSource,
    CinderVolumeSource as _ModelCinderVolumeSource,
    ConfigMapVolumeSource as _ModelConfigMapVolumeSource,
    CSIVolumeSource as _ModelCSIVolumeSource,
    DownwardAPIVolumeSource as _ModelDownwardAPIVolumeSource,
    EmptyDirVolumeSource as _ModelEmptyDirVolumeSource,
    EphemeralVolumeSource as _ModelEphemeralVolumeSource,
    FCVolumeSource as _ModelFCVolumeSource,
    FlexVolumeSource as _ModelFlexVolumeSource,
    FlockerVolumeSource as _ModelFlockerVolumeSource,
    GCEPersistentDiskVolumeSource as _ModelGCEPersistentDiskVolumeSource,
    GitRepoVolumeSource as _ModelGitRepoVolumeSource,
    GlusterfsVolumeSource as _ModelGlusterfsVolumeSource,
    HostPathVolumeSource as _ModelHostPathVolumeSource,
    ISCSIVolumeSource as _ModelISCSIVolumeSource,
    NFSVolumeSource as _ModelNFSVolumeSource,
    ObjectMeta,
    PersistentVolumeClaim as _ModelPersistentVolumeClaim,
    PersistentVolumeClaimSpec as _ModelPersistentVolumeClaimSpec,
    PersistentVolumeClaimVolumeSource as _ModelPersistentVolumeClaimVolumeSource,
    PhotonPersistentDiskVolumeSource as _ModelPhotonPersistentDiskVolumeSource,
    PortworxVolumeSource as _ModelPortworxVolumeSource,
    ProjectedVolumeSource as _ModelProjectedVolumeSource,
    QuobyteVolumeSource as _ModelQuobyteVolumeSource,
    RBDVolumeSource as _ModelRBDVolumeSource,
    ResourceRequirements,
    ScaleIOVolumeSource as _ModelScaleIOVolumeSource,
    SecretVolumeSource as _ModelSecretVolumeSource,
    StorageOSVolumeSource as _ModelStorageOSVolumeSource,
    Volume as _ModelVolume,
    VolumeMount as _ModelVolumeMount,
    VsphereVirtualDiskVolumeSource as _ModelVsphereVirtualDiskVolumeSource,
)
from hera.workflows.validators import validate_storage_units


class AccessMode(Enum):
    """A representations of the volume access modes for Kubernetes.

    Notes
    -----
    See: https://kubernetes.io/docs/concepts/storage/persistent-volumes/#access-modes for more information.
    """

    read_write_once = "ReadWriteOnce"
    """
    The volume can be mounted as read-write by a single node. ReadWriteOnce access mode still can allow multiple
    pods to access the volume when the pods are running on the same node
    """

    read_only_many = "ReadOnlyMany"
    """The volume can be mounted as read-only by many nodes"""

    read_write_many = "ReadWriteMany"
    """The volume can be mounted as read-write by many nodes"""

    read_write_once_pod = "ReadWriteOncePod"
    """
    The volume can be mounted as read-write by a single Pod. Use ReadWriteOncePod access mode if you want to
    ensure that only one pod across whole cluster can read that PVC or write to it. This is only supported for CSI
    volumes and Kubernetes version 1.22+.
    """

    def __str__(self):
        return str(self.value)


class _BaseVolume(_ModelVolumeMount):
    name: Optional[str] = None  # type: ignore

    @validator("name", pre=True)
    def _check_name(cls, v):
        if v is None:
            return str(uuid.uuid4())
        return v

    def _build_persistent_volume_claim(self) -> _ModelPersistentVolumeClaim:
        raise NotImplementedError

    def _build_volume(self) -> _ModelVolume:
        raise NotImplementedError

    def _build_volume_mount(self) -> _ModelVolumeMount:
        return _ModelVolumeMount(
            name=self.name,
            mount_path=self.mount_path,
            mount_propagation=self.mount_propagation,
            read_only=self.read_only,
            sub_path=self.sub_path,
            sub_path_expr=self.sub_path_expr,
        )


class AWSElasticBlockStoreVolumeVolume(_BaseVolume, _ModelAWSElasticBlockStoreVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            aws_elastic_block_store=_ModelAWSElasticBlockStoreVolumeSource(
                fs_type=self.fs_type, partition=self.partition, read_only=self.read_only, volume_id=self.volume_id
            ),
        )


class AzureDiskVolumeVolume(_BaseVolume, _ModelAzureDiskVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            azure_disk=_ModelAzureDiskVolumeSource(
                caching_mode=self.caching_mode,
                disk_name=self.disk_name,
                disk_uri=self.disk_uri,
                fs_type=self.fs_type,
                kind=self.kind,
                read_only=self.read_only,
            ),
        )


class AzureFileVolumeVolume(_BaseVolume, _ModelAzureFileVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            azure_file=_ModelAzureFileVolumeSource(
                read_only=self.read_only, secret_name=self.secret_name, share_name=self.share_name
            ),
        )


class CephFSVolumeVolume(_BaseVolume, _ModelCephFSVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            cephfs=_ModelCephFSVolumeSource(
                monitors=self.monitors,
                path=self.path,
                read_only=self.read_only,
                secret_file=self.secret_file,
                secret_ref=self.secret_ref,
                user=self.user,
            ),
        )


class CinderVolume(_BaseVolume, _ModelCinderVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            cinder=_ModelCinderVolumeSource(
                fs_type=self.fs_type,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
                volume_id=self.volume_id,
            ),
        )


class ConfigMapVolume(_BaseVolume, _ModelConfigMapVolumeSource):  # type: ignore
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            config_map=_ModelConfigMapVolumeSource(
                default_mode=self.default_mode, items=self.items, name=self.name, optional=self.optional
            ),
        )


class CSIVolume(_BaseVolume, _ModelCSIVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            csi=_ModelCSIVolumeSource(
                driver=self.driver,
                fs_type=self.fs_type,
                node_publish_secret_ref=self.node_publish_secret_ref,
                read_only=self.read_only,
                volume_attributes=self.volume_attributes,
            ),
        )


class DownwardAPIVolume(_BaseVolume, _ModelDownwardAPIVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            downward_api=_ModelDownwardAPIVolumeSource(default_mode=self.default_mode, items=self.items),
        )


class EmptyDirVolume(_BaseVolume, _ModelEmptyDirVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name, empty_dir=_ModelEmptyDirVolumeSource(medium=self.medium, size_limit=self.size_limit)
        )


class EphemeralVolume(_BaseVolume, _ModelEphemeralVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name, ephemeral=_ModelEphemeralVolumeSource(volume_claim_template=self.volume_claim_template)
        )


class FCVolume(_BaseVolume, _ModelFCVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            fc=_ModelFCVolumeSource(
                fs_type=self.fs_type,
                lun=self.lun,
                read_only=self.read_only,
                target_ww_ns=self.target_ww_ns,
                wwids=self.wwids,
            ),
        )


class FlexVolume(_BaseVolume, _ModelFlexVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            flex_volume=_ModelFlexVolumeSource(
                driver=self.driver,
                fs_type=self.fs_type,
                options=self.options,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
            ),
        )


class FlockerVolume(_BaseVolume, _ModelFlockerVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            flocker=_ModelFlockerVolumeSource(dataset_name=self.dataset_name, dataset_uuid=self.dataset_uuid),
        )


class GCEPersistentDiskVolume(_BaseVolume, _ModelGCEPersistentDiskVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            gce_persistent_disk=_ModelGCEPersistentDiskVolumeSource(
                fs_type=self.fs_type, partition=self.partition, pd_name=self.pd_name, read_only=self.read_only
            ),
        )


class GitRepoVolume(_BaseVolume, _ModelGitRepoVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            git_repo=_ModelGitRepoVolumeSource(
                directory=self.directory, repository=self.repository, revision=self.revision
            ),
        )


class GlusterfsVolume(_BaseVolume, _ModelGlusterfsVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            glusterfs=_ModelGlusterfsVolumeSource(endpoints=self.endpoints, path=self.path, read_only=self.read_only),
        )


class HostPathVolume(_BaseVolume, _ModelHostPathVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(name=self.name, host_path=_ModelHostPathVolumeSource(path=self.path, type=self.type))


class ISCSIVolume(_BaseVolume, _ModelISCSIVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            iscsi=_ModelISCSIVolumeSource(
                chap_auth_discovery=self.chap_auth_discovery,
                chap_auth_session=self.chap_auth_discovery,
                fs_type=self.fs_type,
                initiator_name=self.initiator_name,
                iqn=self.iqn,
                iscsi_interface=self.iscsi_interface,
                lun=self.lun,
                portals=self.portals,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
                target_portal=self.target_portal,
            ),
        )


class NFSVolume(_BaseVolume, _ModelNFSVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(name=self.name, nfs=_ModelNFSVolumeSource(path=self.path, read_only=self.read_only))


class PhotonPersistentDiskVolume(_BaseVolume, _ModelPhotonPersistentDiskVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            photon_persistent_disk=_ModelPhotonPersistentDiskVolumeSource(fs_type=self.fs_type, pd_id=self.pd_id),
        )


class PortworxVolume(_BaseVolume, _ModelPortworxVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            portworx_volume=_ModelPortworxVolumeSource(
                fs_type=self.fs_type, read_only=self.read_only, volume_id=self.volume_id
            ),
        )


class ProjectedVolume(_BaseVolume, _ModelProjectedVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name, projected=_ModelProjectedVolumeSource(default_mode=self.default_mode, sources=self.sources)
        )


class QuobyteVolume(_BaseVolume, _ModelQuobyteVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            quobyte=_ModelQuobyteVolumeSource(
                group=self.group,
                read_only=self.read_only,
                registry=self.registry,
                tenant=self.tenant,
                user=self.user,
                volume=self.volume,
            ),
        )


class RBDVolume(_BaseVolume, _ModelRBDVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            rbd=_ModelRBDVolumeSource(
                fs_type=self.fs_type,
                image=self.image,
                keyring=self.keyring,
                monitors=self.monitors,
                pool=self.pool,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
                user=self.user,
            ),
        )


class ScaleIOVolume(_BaseVolume, _ModelScaleIOVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            scale_io=_ModelScaleIOVolumeSource(
                fs_type=self.fs_type,
                gateway=self.gateway,
                protection_domain=self.protection_domain,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
                ssl_enabled=self.ssl_enabled,
                storage_mode=self.storage_mode,
                storage_pool=self.storage_pool,
                system=self.system,
                volume_name=self.volume_name,
            ),
        )


class SecretVolume(_BaseVolume, _ModelSecretVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            secret=_ModelSecretVolumeSource(
                default_mode=self.default_mode, items=self.items, optional=self.optional, secret_name=self.secret_name
            ),
        )


class StorageOSVolume(_BaseVolume, _ModelStorageOSVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            storageos=_ModelStorageOSVolumeSource(
                fs_type=self.fs_type,
                read_only=self.read_only,
                secret_ref=self.secret_ref,
                volume_name=self.volume_name,
                volume_namespace=self.volume_namespace,
            ),
        )


class VsphereVirtualDiskVolume(_BaseVolume, _ModelVsphereVirtualDiskVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            vsphere_volume=_ModelVsphereVirtualDiskVolumeSource(
                fs_type=self.fs_type,
                storage_policy_id=self.storage_policy_id,
                storage_policy_name=self.storage_policy_name,
                volume_path=self.volume_path,
            ),
        )


class ExistingVolume(_BaseVolume, _ModelPersistentVolumeClaimVolumeSource):
    def _build_volume(self) -> _ModelVolume:
        return _ModelVolume(
            name=self.name,
            persistent_volume_claim=_ModelPersistentVolumeClaimVolumeSource(
                claim_name=self.claim_name, read_only=self.read_only
            ),
        )


class Volume(_BaseVolume, _ModelPersistentVolumeClaimSpec):
    size: Optional[str] = None  # type: ignore
    resources: Optional[ResourceRequirements] = None
    metadata: Optional[ObjectMeta] = None
    access_modes: Optional[List[Union[str, AccessMode]]] = [AccessMode.read_write_once]  # type: ignore
    storage_class_name: Optional[str] = None

    @validator("access_modes", pre=True, always=True)
    def _check_access_modes(cls, v):
        if not v:
            return [AccessMode.read_write_once]

        result = []
        for mode in v:
            if isinstance(mode, AccessMode):
                result.append(mode)
            else:
                result.append(AccessMode(mode))
        return result

    @validator("name", pre=True, always=True)
    def _check_name(cls, v):
        return v or str(uuid.uuid4())

    @root_validator(pre=True)
    def _merge_reqs(cls, values):
        if "size" in values and "resources" in values:
            resources: ResourceRequirements = values.get("resources")
            if resources.requests is not None:
                if "storage" in resources.requests:
                    pass  # take the storage specification in resources
                else:
                    resources.requests["storage"] = values.get("size")
        elif "resources" not in values:
            assert "size" in values, "at least one of `size` or `resources` must be specified"
            validate_storage_units(cast(str, values.get("size")))
            values["resources"] = ResourceRequirements(requests={"storage": values.get("size")})
        elif "resources" in values:
            resources = cast(ResourceRequirements, values.get("resources"))
            assert resources.requests is not None, "Resource requests are required"
            storage = resources.requests.get("storage")
            assert storage is not None, "At least one of `size` or `resources.requests.storage` must be specified"
            validate_storage_units(cast(str, storage))
        return values

    def _build_persistent_volume_claim(self) -> _ModelPersistentVolumeClaim:
        return _ModelPersistentVolumeClaim(
            metadata=self.metadata or ObjectMeta(name=self.name),
            spec=_ModelPersistentVolumeClaimSpec(
                access_modes=[str(cast(AccessMode, am).value) for am in self.access_modes]
                if self.access_modes is not None
                else None,
                data_source=self.data_source,
                data_source_ref=self.data_source_ref,
                resources=self.resources,
                selector=self.selector,
                storage_class_name=self.storage_class_name,
                volume_mode=self.volume_mode,
                volume_name=self.volume_name,
            ),
        )

    def _build_volume(self) -> _ModelVolume:
        claim = self._build_persistent_volume_claim()
        assert claim.metadata is not None, "claim metadata is required"
        return _ModelVolume(
            name=self.name,
            persistent_volume_claim=_ModelPersistentVolumeClaimVolumeSource(
                claim_name=cast(str, claim.metadata.name),
                read_only=self.read_only,
            ),
        )


__all__ = [
    "AccessMode",
    *[c.__name__ for c in _BaseVolume.__subclasses__()],
]



from typing import Optional

from hera.shared._base_model import BaseModel as BaseModel

from .io.argoproj.events import v1alpha1 as v1alpha1
from .io.k8s.apimachinery.pkg.apis.meta import v1 as v1

class DeleteSensorResponse(BaseModel): ...

class LogEntry(BaseModel):
    dependency_name: Optional[str]
    event_context: Optional[str]
    level: Optional[str]
    msg: Optional[str]
    namespace: Optional[str]
    sensor_name: Optional[str]
    time: Optional[v1.Time]
    trigger_name: Optional[str]

class CreateSensorRequest(BaseModel):
    create_options: Optional[v1.CreateOptions]
    namespace: Optional[str]
    sensor: Optional[v1alpha1.Sensor]

class SensorWatchEvent(BaseModel):
    object: Optional[v1alpha1.Sensor]
    type: Optional[str]

class UpdateSensorRequest(BaseModel):
    name: Optional[str]
    namespace: Optional[str]
    sensor: Optional[v1alpha1.Sensor]



import threading
from typing import List, TypeVar, Union

from hera.shared._base_model import BaseMixin
from hera.workflows.exceptions import InvalidType
from hera.workflows.protocol import Subbable, TTemplate

TNode = TypeVar("TNode", bound="SubNodeMixin")


class SubNodeMixin(BaseMixin):
    """SubNodeMixin ensures that the class gets added to the Hera context on initialization."""

    def __hera_init__(self: TNode) -> TNode:
        _context.add_sub_node(self)
        return self


class _HeraContext(threading.local):
    def __init__(self) -> None:
        super().__init__()
        self._pieces: List[Subbable] = []

    def enter(self, p: Subbable) -> None:
        if not isinstance(p, Subbable):
            raise InvalidType(type(p))
        self._pieces.append(p)

    def exit(self) -> None:
        self._pieces.pop()

    def add_sub_node(self, node: Union[SubNodeMixin, TTemplate]) -> None:
        if self._pieces:
            try:
                # here, we are trying to add a node to the last piece of context in the hopes that it is a subbable
                self._pieces[-1]._add_sub(node)
            except InvalidType:
                # if the above fails, it means the user invoked a decorated function e.g. `@script`. Hence,
                # the object needs to be added as a template to the piece of context at [-1]. This will be the case for
                # DAGs and Steps
                self._pieces[-1]._add_sub(node.template)  # type: ignore

            # when the above does not raise an exception, it means the user invoked a decorated function e.g. `@script`
            # inside a proper context. Here, we add the object to the overall workflow context, directly as a template,
            # in case it is not found (based on the name)
            if hasattr(node, "template") and node.template is not None and not isinstance(node.template, str):
                found = False
                for t in self._pieces[0].templates:  # type: ignore
                    if t.name == node.template.name:
                        found = True
                        break
                if not found:
                    self._pieces[0]._add_sub(node.template)


_context = _HeraContext()



from __future__ import annotations

import inspect
from typing import Callable, Dict, List, Optional, Type, Union

from ._base_model import TBase

Hook = Callable[[TBase], TBase]
_HookMap = Dict[Type[TBase], List[Hook]]


class _GlobalConfig:
    """Hera global configuration holds any user configuration such as global tokens, hooks, etc.

    Notes:
        This should not be instantiated directly by the user. There is an instance of the `_GlobalConfig` in this module,
        which is what should be used. Access as `hera.shared.global_config` or `hera.shared.GlobalConfig`.
    """

    # protected attributes are ones that are computed/go through some light processing upon setting or
    # are processed upon accessing. The rest, which use primitive types, such as `str`, can remain public
    _token: Union[Optional[str], Callable[[], Optional[str]]] = None

    host: Optional[str] = None
    verify_ssl: bool = True
    api_version: str = "argoproj.io/v1alpha1"
    namespace: Optional[str] = None
    _image: Union[str, Callable[[], str]] = "python:3.7"
    service_account_name: Optional[str] = None
    script_command: Optional[List[str]] = ["python"]
    _pre_build_hooks: Optional[_HookMap] = None

    def reset(self) -> None:
        """Resets the global config container to its initial state"""
        self.__dict__.clear()  # Wipe instance values to fallback to the class defaults

    @property
    def image(self) -> str:
        """Return the default image to use for Tasks"""
        if isinstance(self._image, str):
            return self._image
        return self._image()

    @image.setter
    def image(self, image: Union[str, Callable[[], str]]) -> None:
        """Set the default image to use for Tasks"""
        self._image = image

    @property
    def token(self) -> Optional[str]:
        """Returns an Argo Workflows global token"""
        if self._token is None or isinstance(self._token, str):
            return self._token
        return self._token()

    @token.setter
    def token(self, t: Union[Optional[str], Callable[[], Optional[str]]]) -> None:
        """Sets the Argo Workflows token at a global level so services can use it"""
        self._token = t

    def register_pre_build_hook(self, hook: Hook) -> Hook:
        """Registers a hook to be called before building a model."""
        return_type = inspect.signature(hook).return_annotation
        if return_type is inspect.Signature.empty:
            raise TypeError("Hook must have a return type annotation")
        self._pre_build_hooks = self._pre_build_hooks or {}
        self._pre_build_hooks.setdefault(return_type, []).append(hook)
        return hook

    def _get_pre_build_hooks(self, instance: TBase) -> List[Hook]:
        """Registers a hook to be called before building a model."""
        self._pre_build_hooks = self._pre_build_hooks or {}
        return self._pre_build_hooks.get(type(instance)) or []


GlobalConfig = global_config = _GlobalConfig()
register_pre_build_hook = global_config.register_pre_build_hook



from typing import Optional

from hera.shared._base_model import BaseModel as BaseModel

from .io.argoproj.events import v1alpha1 as v1alpha1
from .io.k8s.apimachinery.pkg.apis.meta import v1 as v1

class EventSourceDeletedResponse(BaseModel): ...

class LogEntry(BaseModel):
    event_name: Optional[str]
    event_source_name: Optional[str]
    event_source_type: Optional[str]
    level: Optional[str]
    msg: Optional[str]
    namespace: Optional[str]
    time: Optional[v1.Time]

class CreateEventSourceRequest(BaseModel):
    event_source: Optional[v1alpha1.EventSource]
    namespace: Optional[str]

class EventSourceWatchEvent(BaseModel):
    object: Optional[v1alpha1.EventSource]
    type: Optional[str]

class UpdateEventSourceRequest(BaseModel):
    event_source: Optional[v1alpha1.EventSource]
    name: Optional[str]
    namespace: Optional[str]
